import os

CONFIG = {'local_data_dir': 'goes_data/',
          'remote_data_dir': 'https://satdat.ngdc.noaa.gov/sem/goes/data/'}

# override local data directory with environment variables
if os.environ.get('SPEDAS_DATA_DIR'):
    CONFIG['local_data_dir'] = os.sep.join([os.environ['SPEDAS_DATA_DIR'], 'goes'])

if os.environ.get('GOES_DATA_DIR'):
    CONFIG['local_data_dir'] = os.environ['GOES_DATA_DIR']


"""Tests of cal_fit function."""
import pyspedas.themis.spacecraft.fields.fit
import pytplot.get_data
from pytplot.importers.tplot_restore import tplot_restore
import unittest
import numpy as np
from pyspedas.themis.spacecraft.fields.fit import cal_fit


class TestCalFitDataValidation(unittest.TestCase):
    """Tests of the data been identical to SPEDAS (IDL)."""

    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The IDL script that creates data file:
        https://github.com/spedas/pyspedas-validation/blob/cal_fit/src/themis/validation_files/thm_load_fit_validation_files.pro
        """
        from pyspedas.utilities.download import download
        from pyspedas.themis.config import CONFIG

        # Testing time range
        cls.t = ['2008-03-15', '2008-03-16']

        # Testing tollerange
        cls.tol = 1e-10

        # Download tplot files
        remote_server = 'https://spedas.org/'
        remote_name = 'testfiles/cal_fit.tplot'
        calfile = download(remote_file=remote_name,
                           remote_path=remote_server,
                           local_path=CONFIG['local_data_dir'],
                           no_download=False)
        if not calfile:
            # Skip tests
            raise unittest.SkipTest("Cannot download data validation file")

        # Load validation variables from the test file
        filename = calfile[0]
        tplot_restore(filename)
        cls.tha_fit = pytplot.get_data('tha_fit')
        cls.tha_fgs = pytplot.get_data('tha_fgs')
        cls.tha_fgs_sigma = pytplot.get_data('tha_fgs_sigma')
        cls.tha_fit_bfit = pytplot.get_data('tha_fit_bfit')
        cls.tha_fit_efit = pytplot.get_data('tha_fit_efit')
        cls.tha_efs = pytplot.get_data('tha_efs')
        cls.tha_efs_sigma = pytplot.get_data('tha_efs_sigma')
        cls.tha_efs_0 = pytplot.get_data('tha_efs_0')
        cls.tha_efs_dot0 = pytplot.get_data('tha_efs_dot0')

    def setUp(self):
        """ We need to clean tplot variables before each run"""
        pytplot.del_data('*')

    def test_fgs(self):
        """Validate tha_fgs."""
        pyspedas.themis.fit(trange=self.t, probe='a', level='l1', varnames=['tha_fit'], time_clip=True)

        cal_fit(probe='a')
        tha_fgs = pytplot.get_data('tha_fgs')

        diff = np.nanmedian(tha_fgs.y - self.tha_fgs.y, axis=0, keepdims=True)
        self.assertAlmostEqual(diff.sum(), 0)

    def test_fgs_sigma(self):
        """Validate tha_fgs_sigma."""
        pyspedas.themis.fit(trange=self.t, probe='a', level='l1', varnames=['tha_fit'], time_clip=True)
        cal_fit(probe='a')
        tha_fgs_sigma = pytplot.get_data('tha_fgs_sigma')
        diff = np.nanmedian(tha_fgs_sigma.y - self.tha_fgs_sigma.y, axis=0, keepdims=True)
        self.assertAlmostEqual(diff.sum(), 0)

    def test_fit_bfit(self):
        """Validate tha_fit_bfit."""
        pyspedas.themis.fit(trange=self.t, probe='a', level='l1', varnames=['tha_fit'], time_clip=True)
        cal_fit(probe='a')
        tha_fit_bfit = pytplot.get_data('tha_fit_bfit')
        diff = np.nanmedian(tha_fit_bfit.y - self.tha_fit_bfit.y, axis=0, keepdims=True)
        self.assertAlmostEqual(diff.sum(), 0)

    def test_fit_efit(self):
        """Validate tha_fit_efit."""
        pyspedas.themis.fit(trange=self.t, probe='a', level='l1', varnames=['tha_fit', 'tha_fit_code'],
                            get_support_data=True, time_clip=True)
        cal_fit(probe='a')
        tha_fit_efit = pytplot.get_data('tha_fit_efit')
        diff = np.nanmedian(tha_fit_efit.y - self.tha_fit_efit.y, axis=0, keepdims=True)
        self.assertAlmostEqual(diff.sum(), 0)

    def test_efs(self):
        """Validate tha_efs."""
        pyspedas.themis.fit(trange=self.t, probe='a', level='l1', varnames=['tha_fit', 'tha_fit_code'],
                            get_support_data=True, time_clip=True)
        cal_fit(probe='a')
        tha_efs = pytplot.get_data('tha_efs')
        diff = np.nanmedian(tha_efs.y - self.tha_efs.y, axis=0, keepdims=True)
        self.assertAlmostEqual(diff.sum(), 0)

    def test_efs_sigma(self):
        """Validate tha_efs_sigma."""
        pyspedas.themis.fit(trange=self.t, probe='a', level='l1', varnames=['tha_fit', 'tha_fit_code'],
                            get_support_data=True, time_clip=True)
        cal_fit(probe='a')
        tha_efs_sigma = pytplot.get_data('tha_efs_sigma')
        diff = np.nanmedian(tha_efs_sigma.y - self.tha_efs_sigma.y, axis=0, keepdims=True)
        self.assertAlmostEqual(diff.sum(), 0)

    def test_efs_0(self):
        """Validate tha_efs_0."""
        pyspedas.themis.fit(trange=self.t, probe='a', level='l1', varnames=['tha_fit', 'tha_fit_code'],
                            get_support_data=True, time_clip=True)
        cal_fit(probe='a')
        tha_efs_0 = pytplot.get_data('tha_efs_0')
        diff = np.nanmedian(tha_efs_0.y - self.tha_efs_0.y, axis=0, keepdims=True)
        self.assertAlmostEqual(diff.sum(), 0)

    def test_efs_dot0(self):
        """Validate tha_efs_sigma."""
        pyspedas.themis.fit(trange=self.t, probe='a', level='l1', varnames=['tha_fit', 'tha_fit_code'],
                            get_support_data=True, time_clip=True)
        cal_fit(probe='a')
        tha_efs_dot0 = pytplot.get_data('tha_efs_dot0')
        diff = np.nanmedian(tha_efs_dot0.y - self.tha_efs_dot0.y, axis=0, keepdims=True)
        self.assertAlmostEqual(diff.sum(), 0)


class TestCalFitEfsNoCalDataValidation(unittest.TestCase):
    """Tests of the data been identical to SPEDAS (IDL)."""

    def setUp(self):
        """ IDL Data has to be downloaded to perform these tests"""
        from pyspedas.utilities.download import download
        from pyspedas.themis.config import CONFIG

        # Testing time range
        self.t = ['2008-03-15', '2008-03-16']

        # Testing tollerange
        self.tol = 1e-10

        # Download tplot files
        remote_server = 'https://spedas.org/'
        remote_name = 'testfiles/tha_efs_no_cal.tplot'
        calfile = download(remote_file=remote_name,
                           remote_path=remote_server,
                           local_path=CONFIG['local_data_dir'],
                           no_download=False)
        if not calfile:
            # Skip tests
            self.skipTest("Cannot download data validation file")

        # Load validation variables from the test file
        filename = calfile[0]
        tplot_restore(filename)
        self.tha_efs = pytplot.get_data('tha_efs')

        pytplot.del_data('*')

    def test_efs(self):
        """Validate tha_efs."""
        pyspedas.themis.fit(trange=self.t, probe='a', level='l1', varnames=['tha_fit', 'tha_fit_code'],
                            get_support_data=True, time_clip=True)
        cal_fit(probe='a', no_cal=True)
        tha_efs = pytplot.get_data('tha_efs')
        diff = np.nanmedian(tha_efs.y - self.tha_efs.y, axis=0, keepdims=True)
        self.assertAlmostEqual(diff.sum(), 0)


class TestCalFitInput(unittest.TestCase):

    def setUp(self):
        self.t = ['2008-03-15', '2008-03-15']

    def test_wrong_satellite(self):
        """Validate tha_fgs."""
        pyspedas.themis.fit(trange=self.t, probe='x', level='l1', varnames=['tha_fit'], time_clip=True)
        cal_fit(probe='x')

    def test_e34_ss(self):
        t = ['2011-02-27', '2011-02-28']
        pyspedas.themis.fit(trange=t, probe='b', level='l1', varnames=['thb_fit', 'thb_fit_code'],
                            get_support_data=True, time_clip=True)
        cal_fit(probe='b')

    def test_no_input(self):
        cal_fit(probe="")

    def test_probe_f(self):
        cal_fit(probe="f")


class TestCalFitMeta(unittest.TestCase):
    def setUp(self):
        trange = ['2008-03-15', '2008-03-15']
        pyspedas.themis.fit(trange=trange, probe='a', level='l1', varnames=['tha_fit', 'tha_fit_code'],
                            get_support_data=True, time_clip=True)
        cal_fit(probe='a', no_cal=True)

    def test_meta_units(self):
        vars = {'tha_fgs', 'tha_fgs_sigma', 'tha_fit_bfit', 'tha_fit_efit', 'tha_efs', 'tha_efs_sigma', 'tha_efs_0', 'tha_efs_dot0'}
        for var in vars:
            with self.subTest(f'Test of units in {var}', var=var):
                meta = pytplot.get_data(var, metadata=True)
                self.assertIn('units', meta)
                # axis_subtitle currently displays units
                self.assertIn('axis_subtitle', meta['plot_options']['yaxis_opt'])

    def test_meta_legend(self):
        vars = {'tha_fgs', 'tha_fgs_sigma', 'tha_fit_bfit', 'tha_fit_efit', 'tha_efs', 'tha_efs_sigma', 'tha_efs_0', 'tha_efs_dot0'}
        for var in vars:
            with self.subTest(f'Test of legend in {var}', var=var):
                meta = pytplot.get_data(var, metadata=True)
                self.assertIn('legend_names', meta['plot_options']['yaxis_opt'])


if __name__ == '__main__':
    unittest.main()



import numpy as np


def slice2d_s2c(r, theta, phi):
    """
    Converts spherical coordinates to cartesian

    Input
    ------
        r: array of float
            Radial values

        theta: array of float
            Theta values

        phi: array of float
            Phi values

    Returns
    -------
        Nx3 array of cartesian values (x, y, z)
    """
    rd = np.pi/180.0
    a = np.cos(rd*theta)
    vec = np.array((a*np.cos(rd*phi)*r,
                    a*np.sin(rd*phi)*r,
                    np.sin(rd*theta)*r))
    return vec.T



import logging
from pyspedas import cotrans_get_coord
from pytplot import get_data
from .mms_cotrans_qtransformer import mms_cotrans_qtransformer

try:
    import spacepy.coordinates as coord
except ImportError:
    logging.error('SpacePy must be installed to use this module.')
    logging.error('Please install it using: pip install spacepy')


def mms_qcotrans(in_name=None, out_name=None, in_coord=None, out_coord=None, probe=None):
    """
    Perform coordinate transformations using MMS MEC quaternions.

    This routine uses SpacePy to do the quaternion transformation operations.

    Parameters
    -----------
        in_name: str or list of str
            Names of Tplot variables of vectors to be transformed

        out_name: str or list of str
            Names of output variables

        in_coord: str
            Input coordinate system (e.g. 'bcs','gse','gse2000','gsm','sm','geo','eci')

        out_coord: str
            Output coordinate system (e.g. 'bcs','gse','gse2000','gsm','sm','geo','eci')

        probe: str
            MMS spacecraft # (must be 1, 2, 3, or 4)

    Returns
    --------
        List of variables created

    """
    valid_probes = ['1', '2', '3', '4']
    valid_coords = ['bcs', 'dbcs', 'dmpa', 'smpa', 'dsl', 'ssl', 'gse', 'gse2000', 'gsm', 'sm', 'geo', 'eci', 'j2000']

    if in_name is None:
        logging.error('Input variable name is missing')
        return

    if out_name is None:
        logging.error('Output variable name is missing')
        return

    if not isinstance(in_name, list):
        in_name = [in_name]

    if not isinstance(out_name, list):
        out_name = [out_name]

    out_vars = []

    for idx, variable in enumerate(in_name):
        if in_coord is None:
            var_coords = cotrans_get_coord(variable)
            in_coord = var_coords
            if var_coords is None:
                logging.error('Could not determine coordinate system for: ' + variable)
                continue

        if isinstance(in_coord, list):
            var_coords = in_coord[idx]
        else:
            var_coords = in_coord

        var_coords = var_coords.lower()

        if var_coords not in valid_coords:
            logging.error('Unsupported input coordinate system: ' + var_coords)
            continue

        if isinstance(out_coord, list):
            new_coords = out_coord[idx]
        else:
            new_coords = out_coord

        new_coords = new_coords.lower()

        if new_coords not in valid_coords:
            logging.error('Unsupported output coordinate system: ' + new_coords)

        if var_coords in ['bcs', 'ssl'] or new_coords in ['bcs', 'ssl']:
            logging.warning('WARNING: there are issues transforming data to/from a spinning coordinate system')

        # find the probe, if it's not specified by the user
        if probe is None:
            name_pieces = variable.split('_')
            if len(name_pieces) <= 1:
                logging.warning('Probe could not be determined from: ' + variable + '; defaulting to probe 1')
                probe = '1'
            else:
                probe = name_pieces[0][-1]

        if not isinstance(probe, str):
            probe = str(probe)

        if probe not in valid_probes:
            logging.error('Unknown probe for variable: ' + variable + '; continuing without transforming...')
            continue

        transformed = mms_cotrans_qtransformer(variable, out_name[idx], var_coords, new_coords, probe=probe)

        if transformed is not None:
            out_vars.append(transformed)

            # update the coordinate system in the plot metadata if needed
            metadata = get_data(transformed, metadata=True)
            if metadata.get('plot_options') is not None:
                if metadata['plot_options'].get('yaxis_opt') is not None:
                    if metadata['plot_options']['yaxis_opt'].get('legend_names') is not None:
                        legend = metadata['plot_options']['yaxis_opt'].get('legend_names')
                        updated_legend = [item.replace(var_coords.upper(), new_coords.upper()) for item in legend]
                        metadata['plot_options']['yaxis_opt']['legend_names'] = updated_legend
                    if metadata['plot_options']['yaxis_opt'].get('axis_label') is not None:
                        ytitle = metadata['plot_options']['yaxis_opt'].get('axis_label')
                        metadata['plot_options']['yaxis_opt']['axis_label'] = ytitle.replace(var_coords.upper(),
                                                                                             new_coords.upper())
    return out_vars




import numpy as np
import logging
from pytplot import get_data, store_data, options
from pyspedas.cotrans.cotrans_get_coord import cotrans_get_coord
from pyspedas.cotrans.cotrans import cotrans
from pyspedas.cotrans.gsm2lmn import gsm2lmn
from pyspedas.mms import mec
from pyspedas import tinterpol, omni

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)

def mms_cotrans_lmn(name_in, name_out, gsm=False, gse=False, probe=None, data_rate='srvy'):
    '''
    Tranforms MMS vector fields from GSM coordinates to LMN (boundary-normal) coordinates
    using the Shue et al., 1998 magnetopause model

    Input
    ------
        name_in: str
            Name of the input tplot variable

        name_out: str
            Name of the output tplot variable

    Parameters
    -----------
        gsm: bool
            Flag to indicate the input data are in GSM coordinates

        gse: bool
            Flag to indicate the input data are in GSE coordinates; note: the input data 
            will be transformed to GSM prior to transforming to LMN

        probe: str
            Spacecraft probe #; if not specified, the routine attempts to extract from the 
            input variable name

        data_rate: str
            Data rate of the input data; default is 'srvy'

    Returns
    --------
        Name of the variable containing the data in LMN coordinates.

    '''

    data_in = get_data(name_in)
    metadata_in = get_data(name_in, metadata=True)

    if data_in is None:
        logging.error('Error reading tplot variable: ' + name_in)
        return None

    data_in_coord = cotrans_get_coord(name_in).lower()

    if data_in_coord != 'gse' and data_in_coord != 'gsm' and not gse and not gsm:
        logging.error('Please specify the coordinate system of the input data.')
        return

    # we'll need the probe if it's not specified via keyword
    if probe is None:
        sc_id = name_in.split('_')[0]

        if sc_id != '':
            probe = sc_id[-1]

        if probe not in ['1', '2', '3', '4']:
            logging.error('Error, probe not found; please specify the probe via the "probe" keyword.')
            return

    # load the spacecraft position data
    mec_vars = mec(trange=[min(data_in.times), max(data_in.times)], probe=probe, data_rate=data_rate)

    # interpolate the position data to the input data
    tinterp_vars = tinterpol('mms'+probe+'_mec_r_gsm', name_in, newname='mms'+probe+'_mec_r_gsm_interp')

    pos_data = get_data('mms'+probe+'_mec_r_gsm_interp')

    # we'll need the input data in GSM coordinates
    if data_in_coord != 'gsm':
        cotrans_out = cotrans(name_in, name_in + '_gsm', coord_out='gsm')
        data_in = get_data(name_in + '_gsm')
    
    swdata = solarwind_load([min(data_in.times), max(data_in.times)])

    logging.info('Transforming GSM -> LMN; this may take several minutes, depending on the size of the input.')
    Blmn = gsm2lmn(data_in.times, pos_data.y, data_in.y, swdata=swdata)

    saved = store_data(name_out, data={'x': data_in.times, 'y': Blmn}, attr_dict=metadata_in)

    if saved:
        options(name_out, 'legend_names', ['L', 'M', 'N'])
        return name_out
    else:
        logging.error('Problem creating tplot variable.')

def solarwind_load(trange, level='hro2', min5=False):
    if min5:
        datatype = '5min'
    else:
        datatype = '1min'
    omni_vars = omni.data(trange=trange, level=level, datatype=datatype)
    bzgsm = get_data('BZ_GSM')
    dp = get_data('Pressure')
    return np.array([bzgsm.times, bzgsm.y, dp.y]).T




import logging
from copy import deepcopy
import numpy as np
from pyspedas import time_double
from pytplot import get_data

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def mms_get_fpi_dist(tname, index=None, probe=None, data_rate=None, species=None, level='l2', single_time=None):
    """
    Returns 3D particle data structures containing MMS FPI
    data for use with SPEDAS particle routines. 
    
    Input
    ----------
        tname: str
            tplot variable name containing the FPI distribution data

    Parameters
    ----------
        probe: str
            Spacecraft probe #

        data_rate: str
            Instrument data rates for FPI include 'brst' and 'fast'

        species: str
            Species of the data specified by the 'tname' input
            (valid options: 'i' for ions and 'e' for electrons)

        index: int
            Index of time sample to return

    Returns
    ----------
        3D particle data structure(s) containing MMS FPI distribution functions
    """

    data_in = get_data(tname)

    if data_in is None:
        logging.error('Problem extracting the FPI distribution data.')
        return

    data = [0, 0, 0, 0, 0]

    if single_time is not None:
        index = np.searchsorted(data_in.times, time_double(single_time), side='left')

    if index is not None:
        data[0] = data_in[0][index]
        data[1] = data_in[1][index]
        if data_in[2].ndim == 1:
            data[2] = data_in[2]
        else:
            data[2] = data_in[2][index]
        if data_in[3].ndim == 1:
            data[3] = data_in[3]
        else:
            data[3] = data_in[3][index]
        if data_in[4].ndim == 1:
            data[4] = data_in[4]
        else:
            data[4] = data_in[4][index]
    else:
        data = data_in

    if species is None:
        species = tname.split('_')[1][1]

    if data_rate is None:
        data_rate = tname.split('_')[3]

    if species.lower() == 'i':
        mass = 1.04535e-2
        charge = 1.
        data_name = 'FPI Ion'
        if data_rate is not None and level is not None:
            if data_rate == 'brst' and level == 'l2':
                integ_time = 0.15
            if data_rate == 'brst' and level == 'acr':
                integ_time = 0.0375
        if data_rate == 'fast':
            integ_time = 4.5
    elif species.lower() == 'e':
        mass = 5.68566e-06
        charge = -1.
        data_name = 'FPI Electron'
        if data_rate is not None and level is not None:
            if data_rate == 'brst' and level == 'l2':
                integ_time = 0.03
            if data_rate == 'brst' and level == 'acr':
                integ_time = 0.0075
        if data_rate == 'fast':
            integ_time = 4.5
    else:
        logging.error('Invalid species: ' + species + '; valid options: "i" for ions and "e" for electrons')
        return

    out = {'project_name': 'MMS', 
           'spacecraft': probe, 
           'data_name': data_name, 
           'units_name': 'df_cm', 
           'units_procedure': 'mms_part_conv_units',
           'species': species,
           'valid': 1, # ???
           'charge': charge,
           'mass': mass}

    # the first dimension gets lost on single times
    if isinstance(data[0], np.ndarray) == False:
        data[1] = np.expand_dims(data[1], axis=0)
        data[0] = np.array([data[0]])

    # data[1] shape is: (32, 16, 32) (fast survey data)
    # which is: (phi, theta, energy)

    # we shuffle the output to be [time, energy, phi, theta]
    out_data = data[1].transpose([0, 3, 1, 2])
    out_energy = np.empty(data[1].shape).transpose([0, 3, 1, 2])
    out_theta = np.empty(data[1].shape).transpose([0, 3, 1, 2])
    out_phi = np.empty(data[1].shape).transpose([0, 3, 1, 2])
    out_bins = np.zeros(out_data.shape) + 1
    out_dphi = np.zeros(out_data.shape) + 11.25
    out_dtheta = np.zeros(out_data.shape) + 11.25
    out_denergy = np.zeros(out_data.shape)

    # elevations are constant across time
    # convert colat -> lat
    theta_reform = 90. - np.reshape(data[3], [1, 1, len(data[3])])

    # in the IDL code, we use reform to repeat the vector above
    # here, we'll do the same thing with np.repeat
    theta_rebin1 = np.repeat(theta_reform, data[1].shape[1], axis=0) # repeated across phi
    theta_rebinned = np.repeat(theta_rebin1, data[1].shape[3], axis=1) # repeated across energy

    # energies
    if data[4].ndim == 1:
        energy_reform = np.reshape(data[4], [len(data[4]), 1, 1])
        energy_rebin1 = np.repeat(energy_reform, len(data[3]), axis=2) # repeated across theta
        energy_table = np.repeat(energy_rebin1, len(data[2]), axis=1)
        energy_len = len(data[4])
    elif data[4].ndim == 2:
        energy_len = len(data[4][0])

    # phi
    if data[2].ndim == 1:
        phi_reform = np.reshape(data[2], [1, len(data[2]), 1])
        phi_rebin1 = np.repeat(phi_reform, energy_len, axis=0) # repeated across energy
        phi_rebinned = np.repeat(phi_rebin1, len(data[3]), axis=2)
        phi_len = len(data[2])
    elif data[2].ndim == 2:
        phi_len = len(data[2][0])

    for idx in range(len(data[0])):
        out_theta[idx] = theta_rebinned
        if data[4].ndim == 2: # time varying energy table
            energy_reform = np.reshape(data[4][idx], [energy_len, 1, 1])
            energy_rebin1 = np.repeat(energy_reform, len(data[3]), axis=2) # repeated across theta
            energy_table = np.repeat(energy_rebin1, phi_len, axis=1)
        out_energy[idx] = energy_table
        if data[2].ndim == 2:
            phi_reform = np.reshape(data[2][idx], [1, phi_len, 1])
            phi_rebin1 = np.repeat(phi_reform, energy_len, axis=0) # repeated across energy
            phi_rebinned = np.repeat(phi_rebin1, len(data[3]), axis=2)
        out_phi[idx] = phi_rebinned


    out_phi = (out_phi+180.) % 360
    out_theta = -out_theta

    out_list = []

    for time_idx, time in enumerate(data[0]):
        out_table = {**out}
        out_table['data'] = out_data[time_idx, :]
        out_table['bins'] = out_bins[time_idx, :]
        out_table['theta'] = out_theta[time_idx, :]
        out_table['phi'] = out_phi[time_idx, :]
        out_table['energy'] = out_energy[time_idx, :]
        out_table['dtheta'] = out_dtheta[time_idx, :]
        out_table['dphi'] = out_dphi[time_idx, :]
        out_table['denergy'] = out_denergy[time_idx, :]
        out_table['n_energy'] = energy_len
        out_table['n_theta'] = len(data[3])
        out_table['n_phi'] = phi_len
        out_table['start_time'] = time # note: assumes the FPI data weren't centered!
        out_table['end_time'] = time + integ_time
        out_list.append(out_table)

    return out_list



from copy import deepcopy
import numpy as np

from .slice2d_get_ebounds import slice2d_get_ebounds


def slice2d_get_sphere(dist, energy=False):
    """
    Calculates the center and width of all bins in spherical coordinates.

    Input
    ------
        dist: dict
            3D particle data structure

    Parameters
    -----------
        energy: bool
            Specifies that the user requested energy instead of velocity

    Returns
    --------
        Hash table containing center and width of all bins in spherical coordinates
    """
    n = dist['data'].shape[0]
    c = 299792458.0  # m/s

    # determine gapless energy boundaries.
    ebounds = slice2d_get_ebounds(dist)

    # calculate radial values
    if energy:
        rbounds = ebounds
    else:
        # convert mass from eV/(km/s)^2 to eV/c^2
        erest = dist['mass']*c**2/1e6

        # velocity in km/s (relativistic calc for high energy electrons)
        rbounds = c*np.sqrt(1-1/((ebounds/erest + 1)**2))/1000.0
    
    # get radial centers
    rad = (rbounds[0:n, :, :] + rbounds[1:n+1, :, :])/2.0
    phi = dist['phi'][:, :, :]
    theta = dist['theta'][:, :, :]

    # get bin widths (mainly for geometric method)
    dr = np.abs(rbounds[1:n+1, :, :] - rbounds[0:n, :, :])
    dp = dist['dphi'][:, :, :]
    dt = dist['dtheta'][:, :, :]

    return {'rad': rad,
            'phi': phi,
            'theta': theta,
            'dr': dr,
            'dp': dp,
            'dt': dt}




import os
import unittest
from pyspedas.utilities.data_exists import data_exists

import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_downloadonly(self):
        mag_files = pyspedas.goes.fgm(datatype='1min', downloadonly=True)
        self.assertTrue(os.path.exists(mag_files[0]))

    def test_load_1min_mag_data(self):
        mag_vars = pyspedas.goes.fgm(datatype='1min')
        self.assertTrue(data_exists('BX_1'))
        self.assertTrue(data_exists('BY_1'))
        self.assertTrue(data_exists('BZ_1'))

    def test_load_5min_mag_data(self):
        mag_vars = pyspedas.goes.fgm(datatype='5min')
        self.assertTrue(data_exists('BX_1'))
        self.assertTrue(data_exists('BY_1'))
        self.assertTrue(data_exists('BZ_1'))

    def test_load_mag_data(self):
        mag_vars = pyspedas.goes.fgm(datatype='512ms', suffix='_512')
        self.assertTrue(data_exists('BX_1_512'))
        self.assertTrue(data_exists('BY_1_512'))
        self.assertTrue(data_exists('BZ_1_512'))

    def test_load_1min_epead_data(self):
        epead_vars = pyspedas.goes.epead()
        self.assertTrue(data_exists('E1E_UNCOR_FLUX'))
        self.assertTrue(data_exists('E1W_UNCOR_FLUX'))
        self.assertTrue(data_exists('E2E_UNCOR_FLUX'))

    def test_load_full_epead_data(self):
        epead_vars = pyspedas.goes.epead(datatype='full')
        self.assertTrue(data_exists('E1E_UNCOR_FLUX'))
        self.assertTrue(data_exists('E1W_UNCOR_FLUX'))
        self.assertTrue(data_exists('E2E_UNCOR_FLUX'))

    def test_load_5min_epead_data(self):
        epead_vars = pyspedas.goes.epead(datatype='5min')
        self.assertTrue(data_exists('E1E_UNCOR_FLUX'))
        self.assertTrue(data_exists('E1W_UNCOR_FLUX'))
        self.assertTrue(data_exists('E2E_UNCOR_FLUX'))

    def test_load_full_maged_data(self):
        maged_vars = pyspedas.goes.maged(datatype='full')
        self.assertTrue(data_exists('M_1ME1_DTC_UNCOR_CR'))

    def test_load_1min_maged_data(self):
        maged_vars = pyspedas.goes.maged(time_clip=True)
        self.assertTrue(data_exists('M_1ME1_DTC_UNCOR_FLUX'))

    def test_load_5min_maged_data(self):
        maged_vars = pyspedas.goes.maged(datatype='5min')
        self.assertTrue(data_exists('M_1ME1_DTC_UNCOR_FLUX'))

    def test_load_full_magpd_data(self):
        magpd_vars = pyspedas.goes.magpd(datatype='full')
        self.assertTrue(data_exists('M_1MP1_DTC_UNCOR_CR'))

    def test_load_1min_magpd_data(self):
        magpd_vars = pyspedas.goes.magpd()
        self.assertTrue(data_exists('M_1MP1_DTC_UNCOR_FLUX'))

    def test_load_5min_magpd_data(self):
        magpd_vars = pyspedas.goes.magpd(datatype='5min')
        self.assertTrue(data_exists('M_1MP1_DTC_UNCOR_FLUX'))

    def test_load_full_hepad_data(self):
        hepad_vars = pyspedas.goes.hepad(datatype='full')
        self.assertTrue(data_exists('P10_FLUX'))

    def test_load_1min_hepad_data(self):
        hepad_vars = pyspedas.goes.hepad()
        self.assertTrue(data_exists('P10_FLUX'))

    def test_load_5min_hepad_data(self):
        hepad_vars = pyspedas.goes.hepad(datatype='5min')
        self.assertTrue(data_exists('P10_FLUX'))

    def test_load_xrs_data(self):
        xrs_vars = pyspedas.goes.xrs()
        self.assertTrue(data_exists('A_AVG'))
        xrs_vars = pyspedas.goes.xrs(datatype='5min', trange=['2000-09-01', '2000-09-01'])
        xrs_vars = pyspedas.goes.xrs(datatype='2s')

    def test_load_eps_data(self):
        eps_vars = pyspedas.goes.eps(trange=['2000-09-01', '2000-09-01'], probe='11')
        eps_vars = pyspedas.goes.eps(trange=['2000-09-01', '2000-09-01'], probe='11', datatype='5min')
        self.assertTrue(data_exists('e1_flux_i'))
        self.assertTrue(data_exists('e2_flux_i'))
        self.assertTrue(data_exists('e3_flux_i'))
        self.assertTrue(data_exists('p1_flux'))
        self.assertTrue(data_exists('p2_flux'))
        self.assertTrue(data_exists('p3_flux'))
        self.assertTrue(data_exists('p4_flux'))
        self.assertTrue(data_exists('p5_flux'))
        self.assertTrue(data_exists('p6_flux'))
        self.assertTrue(data_exists('p7_flux'))


if __name__ == '__main__':
    unittest.main()


import unittest
import numpy as np
from pyspedas.mms import mms_load_state, mms_load_mec, mms_load_fgm, mms_load_scm, mms_load_fpi, mms_load_hpca, mms_load_feeps, mms_load_edp, mms_load_edi, mms_load_aspoc, mms_load_dsp
from pyspedas.utilities.data_exists import data_exists
from pyspedas.mms.hpca.mms_hpca_calc_anodes import mms_hpca_calc_anodes
from pyspedas.mms.hpca.mms_hpca_spin_sum import mms_hpca_spin_sum
from pyspedas.mms.hpca.mms_get_hpca_info import mms_get_hpca_info
from pyspedas import tdpwrspc
import pyspedas
from pytplot import get_data, del_data


class FSMLoadTestCases(unittest.TestCase):
    def test_load_basic(self):
        files = pyspedas.mms.fsm(trange=['2015-10-16/06:00', '2015-10-16/06:05'], available=True)
        fsm = pyspedas.mms.fsm(trange=['2015-10-16/06:00', '2015-10-16/06:05'])
        self.assertTrue(data_exists('mms1_fsm_b_mag_brst_l3'))
        self.assertTrue(data_exists('mms1_fsm_b_gse_brst_l3'))


class StateLoadTestCases(unittest.TestCase):
    def test_load_eph_no_update(self):
        data = mms_load_state(datatypes=['pos', 'vel']) # ensure the files are stored locally
        del_data('*') # remove the current tplot vars
        data = mms_load_state(datatypes=['pos', 'vel'], no_update=True) # ensure the files are stored locally
        self.assertTrue(data_exists('mms1_defeph_pos'))
        self.assertTrue(data_exists('mms1_defeph_vel'))

    def test_load_eph_data(self):
        data = mms_load_state(datatypes=['pos', 'vel'])
        self.assertTrue(data_exists('mms1_defeph_pos'))
        self.assertTrue(data_exists('mms1_defeph_vel'))

    def test_load_att_data(self):
        data = mms_load_state(trange=['2015-10-16', '2015-10-16/06:00'], datatypes=['spinras', 'spindec'])
        self.assertTrue(data_exists('mms1_defatt_spinras'))
        self.assertTrue(data_exists('mms1_defatt_spindec'))


############### DSP ############### 
class DSPLoadTestCases(unittest.TestCase):
    # def test_load_default_data(self):
    #     data = mms_load_dsp(trange=['2015-08-01','2015-09-01'], datatype='swd', level='l2', data_rate='fast')
    #     self.assertTrue(data_exists('mms1_dsp_swd_E12_Counts'))

    def test_load_epsd_bpsd_data(self):
        data = mms_load_dsp(trange=['2015-08-01','2015-08-02'], datatype=['epsd', 'bpsd'], level='l2', data_rate='fast')
        self.assertTrue(data_exists('mms1_dsp_epsd_omni'))
        self.assertTrue(data_exists('mms1_dsp_bpsd_omni'))

    def test_load_bpsd_data(self):
        data = mms_load_dsp(trange=['2015-10-16','2015-10-17'], datatype='bpsd', level='l2', data_rate='fast', available=True)
        data = mms_load_dsp(trange=['2015-10-16','2015-10-17'], datatype='bpsd', level='l2', data_rate='fast')
        self.assertTrue(data_exists('mms1_dsp_bpsd_omni_fast_l2'))

    def test_load_epsd_spdf(self):
        data = mms_load_dsp(trange=['2015-08-01','2015-08-02'], datatype='epsd', level='l2', data_rate='fast', spdf=True)
        self.assertTrue(data_exists('mms1_dsp_epsd_omni'))

    def test_load_epsd_suffix(self):
        data = mms_load_dsp(trange=['2015-08-01','2015-08-02'], datatype='epsd', level='l2', data_rate='fast', suffix='_test')
        self.assertTrue(data_exists('mms1_dsp_epsd_omni_test'))


############### FEEPS ############### 
# class FEEPSLoadTestCases(unittest.TestCase):
#     def test_load_default_data(self):
#         data = mms_load_feeps(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
#         data = mms_load_feeps(trange=['2015-10-16', '2015-10-16/01:00'])
#         self.assertTrue(data_exists('mms1_epd_feeps_srvy_l2_electron_intensity_omni'))
#         self.assertTrue(data_exists('mms1_epd_feeps_srvy_l2_electron_intensity_omni_spin'))
#
#     def test_load_spdf_data(self):
#         del_data('*')
#         data = mms_load_feeps(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', spdf=True)
#         self.assertTrue(data_exists('mms1_epd_feeps_brst_l2_electron_intensity_omni'))
#         self.assertTrue(data_exists('mms1_epd_feeps_brst_l2_electron_intensity_omni_spin'))
#
#     def test_load_suffix(self):
#         data = mms_load_feeps(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', suffix='_test')
#         self.assertTrue(data_exists('mms1_epd_feeps_brst_l2_electron_intensity_omni_test'))
#         self.assertTrue(data_exists('mms1_epd_feeps_brst_l2_electron_intensity_omni_spin_test'))
#
#     def test_load_brst_ion_data(self):
#         del_data('*')
#         data = mms_load_feeps(probe=4, data_rate='brst', datatype='ion', trange=['2015-10-01/10:48:16', '2015-10-01/10:49:16'])
#         self.assertTrue(data_exists('mms4_epd_feeps_brst_l2_ion_intensity_omni'))
#         self.assertTrue(data_exists('mms4_epd_feeps_brst_l2_ion_intensity_omni_spin'))
#
#     def test_load_brst_multi_probe(self):
#         del_data('*')
#         data = mms_load_feeps(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', probe=[1, 2, 3, 4])
#         self.assertTrue(data_exists('mms1_epd_feeps_brst_l2_electron_intensity_omni'))
#         self.assertTrue(data_exists('mms1_epd_feeps_brst_l2_electron_intensity_omni_spin'))
#         self.assertTrue(data_exists('mms2_epd_feeps_brst_l2_electron_intensity_omni'))
#         self.assertTrue(data_exists('mms2_epd_feeps_brst_l2_electron_intensity_omni_spin'))
#         self.assertTrue(data_exists('mms3_epd_feeps_brst_l2_electron_intensity_omni'))
#         self.assertTrue(data_exists('mms3_epd_feeps_brst_l2_electron_intensity_omni_spin'))
#         self.assertTrue(data_exists('mms4_epd_feeps_brst_l2_electron_intensity_omni'))
#         self.assertTrue(data_exists('mms4_epd_feeps_brst_l2_electron_intensity_omni_spin'))


############### FPI ############### 
# class FPILoadTestCases(unittest.TestCase):
#     def test_load_default_data(self):
#         data = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'], available=True)
#         data = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'])
#         self.assertTrue(data_exists('mms1_dis_energyspectr_omni_fast'))
#
#     def test_load_spdf_data(self):
#         data = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'], spdf=True)
#         self.assertTrue(data_exists('mms1_dis_energyspectr_omni_fast'))
#
#     def test_load_small_brst_interval(self):
#         data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype=['dis-moms', 'dis-dist'], time_clip=True)
#         self.assertTrue(data_exists('mms1_dis_energyspectr_omni_brst'))
#
#     def test_load_rename_bars(self):
#         data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype='des-dist')
#         data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype='dis-dist')
#         data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype='des-moms')
#         data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype='dis-moms')
#         self.assertTrue(data_exists('mms1_dis_compressionloss_brst_moms'))
#         self.assertTrue(data_exists('mms1_dis_errorflags_brst_moms'))
#         self.assertTrue(data_exists('mms1_des_errorflags_brst_moms'))
#         self.assertTrue(data_exists('mms1_des_compressionloss_brst_moms'))
#         self.assertTrue(data_exists('mms1_des_errorflags_brst_dist'))
#         self.assertTrue(data_exists('mms1_des_compressionloss_brst_dist'))
#         self.assertTrue(data_exists('mms1_dis_errorflags_brst_dist'))
#         self.assertTrue(data_exists('mms1_dis_compressionloss_brst_dist'))
#
#     def test_center_fast_ion_data(self):
#         data = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'])
#         centered = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'], center_measurement=True, suffix='_centered')
#
#         t, d = get_data('mms1_dis_bulkv_gse_fast')
#         c, d = get_data('mms1_dis_bulkv_gse_fast_centered')
#         self.assertTrue(np.round(c[0]-t[0], decimals=3) == 2.25)
#
#     def test_center_fast_electron_data(self):
#         data = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'])
#         centered = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'], center_measurement=True, suffix='_centered')
#
#         t, d = get_data('mms1_des_bulkv_gse_fast')
#         c, d = get_data('mms1_des_bulkv_gse_fast_centered')
#         self.assertTrue(np.round(c[0]-t[0], decimals=3) == 2.25)
#
#     def test_center_brst_ion_data(self):
#         data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst')
#         centered = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', center_measurement=True, suffix='_centered')
#
#         t, d = get_data('mms1_dis_bulkv_gse_brst')
#         c, d = get_data('mms1_dis_bulkv_gse_brst_centered')
#         self.assertTrue(np.round(c[0]-t[0], decimals=3) == 0.075)
#
#     def test_center_brst_electron_data(self):
#         data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst')
#         centered = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', center_measurement=True, suffix='_centered')
#
#         t, d = get_data('mms1_des_bulkv_gse_brst')
#         c, d = get_data('mms1_des_bulkv_gse_brst_centered')
#         self.assertTrue(np.round(c[0]-t[0], decimals=3) == 0.015)
#
#     def test_errorflag_compression_bars(self):
#         data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype=['des-dist', 'des-moms'])
#         data = mms_load_fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', datatype=['dis-dist', 'dis-moms'])
#         # mms_fpi_make_errorflagbars('mms1_des_errorflags_brst_moms', level='l2')
#         # mms_fpi_make_errorflagbars('mms1_dis_errorflags_brst_moms', level='l2')
#         # mms_fpi_make_errorflagbars('mms1_des_errorflags_brst_dist', level='l2')
#         # mms_fpi_make_errorflagbars('mms1_dis_errorflags_brst_dist', level='l2')
#         # mms_fpi_make_compressionlossbars('mms1_des_compressionloss_brst_moms')
#         # mms_fpi_make_compressionlossbars('mms1_dis_compressionloss_brst_moms')
#         # mms_fpi_make_compressionlossbars('mms1_des_compressionloss_brst_dist')
#         # mms_fpi_make_compressionlossbars('mms1_dis_compressionloss_brst_dist')
#         self.assertTrue(data_exists('mms1_des_errorflags_brst_moms_flagbars_full'))
#         self.assertTrue(data_exists('mms1_des_errorflags_brst_moms_flagbars_main'))
#         self.assertTrue(data_exists('mms1_des_errorflags_brst_moms_flagbars_mini'))
#         self.assertTrue(data_exists('mms1_dis_errorflags_brst_moms_flagbars_full'))
#         self.assertTrue(data_exists('mms1_dis_errorflags_brst_moms_flagbars_main'))
#         self.assertTrue(data_exists('mms1_dis_errorflags_brst_moms_flagbars_mini'))
#         self.assertTrue(data_exists('mms1_des_errorflags_brst_dist_flagbars_dist'))
#         self.assertTrue(data_exists('mms1_dis_errorflags_brst_dist_flagbars_dist'))
#         self.assertTrue(data_exists('mms1_des_compressionloss_brst_moms_flagbars'))
#         self.assertTrue(data_exists('mms1_dis_compressionloss_brst_moms_flagbars'))
#         self.assertTrue(data_exists('mms1_des_compressionloss_brst_dist_flagbars'))
#         self.assertTrue(data_exists('mms1_dis_compressionloss_brst_dist_flagbars'))


############### HPCA ############### 
class HPCALoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_hpca(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_hpca(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_hpca_hplus_number_density'))

    def test_load_spdf_data(self):
        data = mms_load_hpca(trange=['2015-10-16', '2015-10-16/01:00'], spdf=True)
        self.assertTrue(data_exists('mms1_hpca_hplus_number_density'))

    def test_load_ion_omni_suffix(self):
        del_data('*')
        data = mms_load_hpca(probe=2, trange=['2016-08-09/09:10', '2016-08-09/10:10:00'], datatype='ion', data_rate='brst', suffix='_brst')
        mms_hpca_calc_anodes(fov=[0, 360], probe=2, suffix='_brst')
        mms_hpca_spin_sum(probe=2, suffix='_brst', avg=True)
        self.assertTrue(data_exists('mms2_hpca_hplus_flux_brst_elev_0-360_spin'))

    def test_load_ion_omni(self):
        del_data('*')
        data = mms_load_hpca(trange=['2016-10-16/5:00', '2016-10-16/6:00'], datatype='ion')
        mms_hpca_calc_anodes(fov=[0, 360], probe='1')
        mms_hpca_spin_sum()
        self.assertTrue(data_exists('mms1_hpca_hplus_flux_elev_0-360_spin'))
        self.assertTrue(data_exists('mms1_hpca_heplus_flux_elev_0-360_spin'))
        self.assertTrue(data_exists('mms1_hpca_heplusplus_flux_elev_0-360_spin'))
        self.assertTrue(data_exists('mms1_hpca_oplus_flux_elev_0-360_spin'))

    def test_center_fast_moments_data(self):
        data = mms_load_hpca(trange=['2015-10-16/14:00', '2015-10-16/15:00'])
        centered = mms_load_hpca(trange=['2015-10-16/14:00', '2015-10-16/15:00'], center_measurement=True, suffix='_centered')
        
        t, d = get_data('mms1_hpca_hplus_ion_bulk_velocity')
        c, d = get_data('mms1_hpca_hplus_ion_bulk_velocity_centered')
        self.assertTrue(np.round(c[0]-t[0], decimals=3) == 5.0)

    def test_center_brst_moments_data(self):
        data = mms_load_hpca(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst')
        centered = mms_load_hpca(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', center_measurement=True, suffix='_centered')
        
        t, d = get_data('mms1_hpca_hplus_ion_bulk_velocity')
        c, d = get_data('mms1_hpca_hplus_ion_bulk_velocity_centered')
        self.assertTrue(np.round(c[0]-t[0], decimals=3) == 5.0)


############### EDI ############### 
class EDILoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_edi(trange=['2016-10-17/13:00', '2016-10-17/14:00'], available=True)
        data = mms_load_edi(trange=['2016-10-17/13:00', '2016-10-17/14:00'])
        self.assertTrue(data_exists('mms1_edi_e_gse_srvy_l2'))

    def test_load_spdf_data(self):
        data = mms_load_edi(trange=['2016-10-17/13:00', '2016-10-17/14:00'], spdf=True)
        self.assertTrue(data_exists('mms1_edi_e_gse_srvy_l2'))

    def test_load_suffix(self):
        data = mms_load_edi(trange=['2016-10-17/13:00', '2016-10-17/14:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_edi_e_gse_srvy_l2_test'))


############### ASPOC ############### 
class ASPOCLoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_aspoc(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_aspoc(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_aspoc_ionc_l2'))

    def test_load_spdf_data(self):
        data = mms_load_aspoc(trange=['2015-10-16', '2015-10-16/01:00'], spdf=True)
        self.assertTrue(data_exists('mms1_aspoc_ionc_l2'))

    def test_load_suffix(self):
        data = mms_load_aspoc(trange=['2015-10-16', '2015-10-16/01:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_aspoc_ionc_l2_test'))


############### EDP ############### 
class EDPLoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_edp_dce_gse_fast_l2'))

    def test_load_hfesp_data(self):
        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'], datatype='hfesp', data_rate='srvy')
        self.assertTrue(data_exists('mms1_edp_hfesp_srvy_l2'))

    def test_load_spdf_data(self):
        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'], spdf=True)
        self.assertTrue(data_exists('mms1_edp_dce_gse_fast_l2'))

    def test_load_suffix(self):
        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_edp_dce_gse_fast_l2'))

    # def test_load_multiple_sc(self):
    #     data = mms_load_edp(probe=['1', '2', '3', '4'], trange=['2015-10-16', '2015-10-16/01:00'])
    #     self.assertTrue(data_exists('mms1_edp_dce_gse_fast_l2'))
    #     self.assertTrue(data_exists('mms2_edp_dce_gse_fast_l2'))
    #     self.assertTrue(data_exists('mms3_edp_dce_gse_fast_l2'))
    #     self.assertTrue(data_exists('mms4_edp_dce_gse_fast_l2'))

    def test_load_brst_data(self):
        data = mms_load_edp(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'])
        self.assertTrue(data_exists('mms1_edp_dce_gse_brst_l2'))


############### FGM ############### 
class FGMLoadTestCases(unittest.TestCase):
    def test_regression_multi_imports_spdf(self):
        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], spdf=True)
        t1, d1 = get_data('mms1_fgm_b_gse_brst_l2')
        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], spdf=True)
        t2, d2 = get_data('mms1_fgm_b_gse_brst_l2')
        self.assertTrue(t1.shape == t2.shape)
        self.assertTrue(d1.shape == d2.shape)

    def test_load_default_data(self):
        data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_fgm_b_gse_srvy_l2'))
        self.assertTrue(data_exists('Epoch'))
        self.assertTrue(data_exists('Epoch_state'))

    def test_load_spdf_data(self):
        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], spdf=True)
        self.assertTrue(data_exists('mms1_fgm_b_gse_brst_l2'))

    def test_load_suffix(self):
        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], suffix='_test')
        self.assertTrue(data_exists('mms1_fgm_b_gse_brst_l2_test'))

    def test_load_multiple_sc(self):
        data = mms_load_fgm(probe=[1, 2, 3, 4], data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'])
        self.assertTrue(data_exists('mms1_fgm_b_gse_brst_l2'))
        self.assertTrue(data_exists('mms2_fgm_b_gse_brst_l2'))
        self.assertTrue(data_exists('mms3_fgm_b_gse_brst_l2'))
        self.assertTrue(data_exists('mms4_fgm_b_gse_brst_l2'))

    def test_load_brst_data(self):
        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'])
        self.assertTrue(data_exists('mms1_fgm_b_gse_brst_l2'))

    def test_load_data_no_update(self):
        data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00']) # make sure the files exist locally
        del_data('*') 
        data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'], no_update=True) # load the file from the local cache
        self.assertTrue(data_exists('mms1_fgm_b_gse_srvy_l2'))


############### MEC ###############
class MECLoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_mec(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_mec(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_mec_r_sm'))

    def test_load_spdf_data(self):
        data = mms_load_mec(trange=['2015-10-16', '2015-10-16/01:00'], spdf=True)
        self.assertTrue(data_exists('mms1_mec_r_sm'))

    def test_load_suffix(self):
        data = mms_load_mec(trange=['2015-10-16', '2015-10-16/01:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_mec_r_sm_test'))


class SCMLoadTestCases(unittest.TestCase):
    def test_brst_dpwrspc_data(self):
        data = mms_load_scm(probe=4, data_rate='brst', datatype='scb', trange=['2015-10-01/10:48:16', '2015-10-01/10:49:16'])
        tdpwrspc('mms4_scm_acb_gse_scb_brst_l2', notmvariance=True)
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_x_dpwrspc'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_y_dpwrspc'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_z_dpwrspc'))

    def test_load_default_data(self):
        data = mms_load_scm(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_scm(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_scm_acb_gse_scsrvy_srvy_l2'))

    def test_load_schb(self):
        data = pyspedas.mms.scm(probe=4, data_rate='brst', datatype='schb', trange=['2015-10-01/10:48:16', '2015-10-01/10:49:16'])
        self.assertTrue(data_exists('mms4_scm_acb_gse_schb_brst_l2'))

    def test_load_suffix(self):
        data = mms_load_scm(trange=['2015-10-16', '2015-10-16/01:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_scm_acb_gse_scsrvy_srvy_l2_test'))

    def test_load_multiple_sc(self):
        data = mms_load_scm(probe=['1', '2', '3', '4'], trange=['2017-12-15', '2017-12-16'])
        # self.assertTrue(data_exists('mms1_scm_acb_gse_scsrvy_srvy_l2'))
        # self.assertTrue(data_exists('mms2_scm_acb_gse_scsrvy_srvy_l2'))
        self.assertTrue(data_exists('mms3_scm_acb_gse_scsrvy_srvy_l2'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scsrvy_srvy_l2'))

    def test_load_brst_data(self):
        data = mms_load_scm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], datatype='scb')
        self.assertTrue(data_exists('mms1_scm_acb_gse_scb_brst_l2'))

    def test_available(self):
        files = mms_load_scm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], datatype='scb', available=True)
        self.assertTrue(len(files) == 2)

    def test_hpca_info(self):
        info = mms_get_hpca_info()
        self.assertTrue(list(info.keys()) == ['elevation', 't_spin', 't_sweep', 'azimuth_energy_offset'])


if __name__ == '__main__':
    unittest.main()



import numpy as np
from pytplot import get_data, store_data


def tnormalize(variable, newname=None, return_data=False):
    """
    Normalize all the vectors stored in a tplot variable

    Input
    ----------
        variable: str or np.ndarray
            tplot variable (or numpy array) containing the vectors to be normalized

    Parameters
    ----------
        newname: str
            name of the output variable; default: variable_normalized

        return_data: bool
            return the normalized vectors instead of creating a tplot variable
    
    Returns
    ----------

        name of the tplot variable created or normalized vectors if return_data
        is set

    """
    metadata_in = {}
    if isinstance(variable, str):
        data_in = get_data(variable)
        metadata_in = get_data(variable, metadata=True)
        data = data_in[1]
        times = data_in[0]
    else:
        data = np.atleast_2d(variable)
        times = np.zeros(data.shape[0])

    n = np.sqrt(np.nansum(data**2, axis=1))

    # to do element-wise division, the magnitude needs to be repeated for each component
    norm_reshaped = np.reshape(n, [len(times), 1])
    norm_mag = np.repeat(norm_reshaped, len(data[0, :]), axis=1)

    data_norm = data/norm_mag

    if return_data:
        return data_norm
    else:
        if newname is None:
            newname = variable + '_normalized'
        store_data(newname, data={'x': times, 'y': data_norm}, attr_dict=metadata_in)
        return newname




import numpy as np
from astropy.coordinates import spherical_to_cartesian, cartesian_to_spherical


def spd_pgs_do_fac(data_in, mat):
    """
    Applies field aligned coordinate transformation to input data

    Input:
        data_in: dict
            Particle data structure to be rotated

        mat: numpy.ndarray
            The 3x3 field-aligned rotation matrix to apply to the data

    Returns:
        Rotated particle data structure
    """

    data_out = data_in.copy()
    rvals = np.ones(data_in['data'].shape)
    cart_data = spherical_to_cartesian(rvals, data_in['theta']*np.pi/180.0, data_in['phi']*np.pi/180.0)

    x = mat[0, 0]*cart_data[0] + mat[0, 1]*cart_data[1] + mat[0, 2]*cart_data[2]
    y = mat[1, 0]*cart_data[0] + mat[1, 1]*cart_data[1] + mat[1, 2]*cart_data[2]
    z = mat[2, 0]*cart_data[0] + mat[2, 1]*cart_data[1] + mat[2, 2]*cart_data[2]

    sphere_data = cartesian_to_spherical(x, y, z)

    data_out['theta'] = sphere_data[1].value*180.0/np.pi
    data_out['phi'] = sphere_data[2].value*180.0/np.pi

    return data_out



from pyspedas.mms.mms_load_data import mms_load_data
from pyspedas.mms.feeps.mms_feeps_correct_energies import mms_feeps_correct_energies
from pyspedas.mms.feeps.mms_feeps_flat_field_corrections import mms_feeps_flat_field_corrections
from pyspedas.mms.feeps.mms_feeps_active_eyes import mms_feeps_active_eyes
from pyspedas.mms.feeps.mms_feeps_split_integral_ch import mms_feeps_split_integral_ch
from pyspedas.mms.feeps.mms_feeps_remove_bad_data import mms_feeps_remove_bad_data
from pyspedas.mms.feeps.mms_feeps_remove_sun import mms_feeps_remove_sun
from pyspedas.mms.feeps.mms_feeps_omni import mms_feeps_omni
from pyspedas.mms.feeps.mms_feeps_spin_avg import mms_feeps_spin_avg
from pyspedas.mms.print_vars import print_vars
from pyspedas.mms.mms_config import CONFIG

@print_vars
def mms_load_feeps(trange=['2015-10-16', '2015-10-17'], probe='1', data_rate='srvy', 
    level='l2', datatype='electron', varformat=None, varnames=[], get_support_data=True, suffix='', time_clip=False,
    no_update=False, available=False, notplot=False, no_flatfield_corrections=False, data_units=['count_rate', 'intensity'], 
    latest_version=False, major_version=False, min_version=None, cdf_version=None, spdf=False, always_prompt=False):
    """
    This function loads FEEPS data into tplot variables
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe : str or list of str
            list of probes, valid values for MMS probes are ['1','2','3','4']. 

        data_rate : str or list of str
            instrument data rates for FEEPS include ['brst', 'srvy']. The
            default is 'srvy'.

        level : str
            indicates level of data processing. the default if no level is specified is 'l2'

        datatype : str or list of str
            Valid datatypes for FEEPS are: 
                       L2, L1b: ['electron', 'ion']
                       L1a: ['electron-bottom', 'electron-top', 'ion-bottom', 'ion-top']

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        time_clip: bool
            Data will be clipped to the exact trange specified by the trange keyword.
            
        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        notplot: bool
            If True, then data are returned in a hash table instead of 
            being stored in tplot variables (useful for debugging, and
            access to multi-dimensional data products)

        available: bool
            If True, simply return the available data files (without downloading)
            for the requested paramters

        no_update: bool
            Set this flag to preserve the original data. if not set and newer 
            data is found the existing data will be overwritten

        cdf_version: str
            Specify a specific CDF version # to load (e.g., cdf_version='4.3.0')

        min_version: str
            Specify a minimum CDF version # to load

        latest_version: bool
            Only grab the latest CDF version in the requested time interval

        major_version: bool
            Only open the latest major CDF version (e.g., X in vX.Y.Z) in the requested time interval

        always_prompt: bool
            Set this keyword to always prompt for the user's username and password;
            useful if you accidently save an incorrect password, or if your SDC password has changed

        spdf: bool
            If True, download the data from the SPDF instead of the SDC

    Returns:
        List of tplot variables created.

    """
    tvars = mms_load_data(trange=trange, notplot=notplot, probe=probe, data_rate=data_rate, level=level, instrument='feeps',
            datatype=datatype, varformat=varformat, varnames=varnames, get_support_data=get_support_data, suffix=suffix,
            time_clip=time_clip, no_update=no_update, available=available, latest_version=latest_version, 
            major_version=major_version, min_version=min_version, cdf_version=cdf_version, spdf=spdf, always_prompt=always_prompt)

    if tvars == [] or available or notplot or CONFIG['download_only']:
        return tvars

    probes = probe if isinstance(probe, list) else [probe]
    data_rates = data_rate if isinstance(data_rate, list) else [data_rate]
    levels = level if isinstance(level, list) else [level]
    datatypes = datatype if isinstance(datatype, list) else [datatype]
    data_units = data_units if isinstance(data_units, list) else [data_units]

    probes = [str(p) for p in probes]

    mms_feeps_correct_energies(probes, data_rate, level=level, suffix=suffix)

    if not no_flatfield_corrections:
        mms_feeps_flat_field_corrections(probes=probes, data_rate=data_rate, suffix=suffix)

    for probe in probes:
        for lvl in levels:
            for drate in data_rates:
                for datatype in datatypes:
                   mms_feeps_remove_bad_data(trange=trange, probe=probe, data_rate=drate, datatype =datatype, level=lvl, suffix=suffix)

                   for data_unit in data_units:
                       eyes = mms_feeps_active_eyes(trange, probe, drate, datatype, lvl)

                       split_vars = mms_feeps_split_integral_ch(data_unit, datatype, probe, suffix=suffix, data_rate=drate, level=lvl, sensor_eyes=eyes)

                       sun_removed_vars = mms_feeps_remove_sun(eyes, trange, probe=probe, datatype=datatype, data_units=data_unit, data_rate=drate, level=lvl, suffix=suffix)

                       omni_vars = mms_feeps_omni(eyes, probe=probe, datatype=datatype, data_units=data_unit, data_rate=drate, level=lvl, suffix=suffix)

                       tvars = tvars + split_vars + sun_removed_vars + omni_vars
                       
                       tvars.append(mms_feeps_spin_avg(probe=probe, data_units=data_unit, datatype=datatype, data_rate=drate, level=lvl, suffix=suffix))

    return tvars



"""
@Author: Xin Cao, Xiangning Chu, University of Colorado Boulder
This version: this function is designed to read EICS or SECS data, and return it as a pandas dataframe.
"""
import os
from .load import load
import numpy as np
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.analysis.time_clip import time_clip as tclip
from pyspedas.utilities.time_double import time_double
import pandas as pd
import time
import zipfile
import pyspedas
import logging
import shutil
import gzip
import pickle
import os


def data(trange=['2012-11-05/00:00:00', '2012-11-06/00:00:00'], resolution=10, dtype=None, no_download = False, downloadonly = False, out_type = 'np', save_pickle = False):
    """
    This function loads SECS/EICS data

    Parameters
    ----------
        trange: list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        resolution: the resolution of the data in seconds, the default is 10s.

        dtype: str
            Data type; valid options: 'EICS' or 'SECS'

        no_download: bool
            Set this flag not to duplicately download the data files.

        downloadonly: bool
            Set this flag only download the data but not read.

        out_type: str
            Set the type of the return/output to be
            'df' (pandas dataframe), 'np' (numpy array) or 'dc' (dictionary).
    Returns
    ----------
        list of str of downloaded filenames (if downloadonly == True)
        or
        the data which is read from the downloaded files..
    """
    return load(trange = trange, resolution=resolution, dtype = dtype, no_download = no_download, downloadonly = downloadonly, out_type = out_type, save_pickle = save_pickle)


def read_data_files(out_files = None, dtype = None, out_type = 'np', save_pickle = False):
    """
    Read data on a daily basis with a 10-secs or other resolution
    :param out_files: the string list of the downloaded data files' path.
    :param out_type: the return type: 'np': numpy array; 'df': pandas dataframe; 'dc': dictionary
    :param dtype: the data which will be read ('EICS' or 'SECS')
    :return: a numpy nd-array acrossing one or multiple days.
    """
    file_names_arr_Dir = out_files
    start_time = time.time()
    # Reading the data at each time stamp (per resolution secs) on one specific date.

    # input the data into one pd data frame. (four columns)
    if out_type == 'df':
        if dtype == 'EICS':
            colnames = ['latitude', 'longitude', 'Jx', 'Jy']
        if dtype == 'SECS':
            colnames = ['latitude', 'longitude', 'J']
        data_all = []
        for idx, file in enumerate(file_names_arr_Dir):
            df = pd.read_csv(file, header=None, sep='\s+', skiprows=0, names=colnames)
            df['datetime'] = file[-19:-4]
            data_all.append(df)
        output = pd.concat(data_all, axis=0, ignore_index=True)

    elif out_type == 'np':
        latitude = []
        longitude = []
        date_time = []
        if dtype == 'EICS':
            Jx = []
            Jy = []
            for file in file_names_arr_Dir:
                di = np.loadtxt(file)
                num_row = np.shape(di)[0]
                latitude.extend(di[:, 0])
                longitude.extend(di[:, 1])
                Jx.extend(di[:, 2])
                Jy.extend(di[:, 3])
                date_time.extend(np.full((num_row, 1), file[-19:-4]))
            num_row2 = len(latitude)
            data_all = np.array([latitude, longitude, Jx, Jy, date_time])
            data_all = data_all.reshape([5, num_row2])
            data_all = np.transpose(data_all)

        if dtype == 'SECS':
            J = []
            for file in file_names_arr_Dir:
                di = np.loadtxt(file)
                num_row = np.shape(di)[0]
                latitude.extend(di[:, 0])
                longitude.extend(di[:, 1])
                J.extend(di[:, 2])
                date_time.extend(np.full((num_row, 1), file[-19:-4]))
            num_row2 = len(latitude)
            data_all = np.array([latitude, longitude, J, date_time])
            data_all = data_all.reshape([4, num_row2])
            data_all = np.transpose(data_all)

        output = data_all

    elif out_type == 'dc':
        data_dict = {}
        Jx = []
        Jy = []
        J = []

        date_time = []
        flag = 0
        filename_day1 = file_names_arr_Dir[0]


        for idx, file in enumerate(file_names_arr_Dir): # per dat file with 1 min resolution.
            if not os.path.isfile(file):
                continue # jump ouf of the current iteration, into the next iteration of the same loop.
            if os.stat(file).st_size == 0: # check if the file is empty.
                continue

            di = np.loadtxt(file)
            if np.shape(di)[0] > 0 and flag == 0:
                num_row = np.shape(di)[0] # np array
                latitude = di[:, 0] # np array
                longitude = di[:, 1] # np array
                flag = 1

            if dtype == 'EICS':
                Jx.append(di[:, 2]) # list [np.arrays]
                Jy.append(di[:, 3]) # list [np.arrays]
            if dtype == 'SECS':
                J.append(di[:, 2])  # list [np.arrays]
            date_time.append(file[-19:-4]) # list of str

        date_time = np.array(date_time) # np array of str
        date_time = time_double(date_time) # np array of float
        if dtype == 'EICS':
            Jx = np.vstack(Jx) # np array
            Jy = np.vstack(Jy) # np array
            data_dict = {'time': date_time, 'latitude': latitude, 'longitude': longitude, 'Jx': Jx, 'Jy': Jy}
        if dtype == 'SECS':
            J = np.vstack(J)  # np array
            data_dict = {'time': date_time, 'latitude': latitude, 'longitude': longitude, 'J': J}
        output = data_dict


    else:
        raise TypeError("%r are invalid keyword arguments" % out_type)

    if save_pickle == True:
        if out_type == 'dc':  # too large, not useful.
            with open('data_dc.pkl', 'wb') as f:
                pickle.dump(output, f)

        # f.close()
    logging.info('running time of output ' + out_type + ": --- %s seconds ---" % (time.time() - start_time))

    return output


def data_selecttime(data = None, dtime = None):
    Data_Days = data
    tp = dtime
    datetime_tp = tp[0:4] + tp[5:7] + tp[8:10] + '_' + tp[11:13] + tp[14:16] + tp[17:19]
    Data_Days_time = Data_Days.loc[Data_Days['datetime'] == datetime_tp]
    return Data_Days_time



